#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module containing the datasets classes source code for :class:`~aac_datasets.datasets.audiocaps.AudioCaps`, :class:`~aac_datasets.datasets.clotho.Clotho`, :class:`~aac_datasets.datasets.macs.MACS` and :class:`~aac_datasets.datasets.wavcaps.WavCaps`.
"""

from .audiocaps import AudioCaps
from .clotho import Clotho
from .macs import MACS
from .wavcaps import WavCaps
